#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include "bgmfile.h"

/* #define DEBUG */

struct BGMmap
{
    char* name;
    int size;
    char* mapped;
};

struct BGM
{
    int n;			/* number of tracks */
    int cur;			/* current track */
    struct BGMmap* tracks;
};

struct BGMfiles
{
    int n;
    char* pathroot;
    struct BGM* bgm;
};

static struct BGM* current_bgm = NULL;
static struct BGMfiles bgm;

void init_bgm(const char* bgmindex)
{
    char buff[BUFSIZ];
    char* p;
    FILE* fp;
    int i;

    strcpy(buff, bgmindex);
    if((p = strrchr(buff, '/')) != NULL)
	*p = '\0';
    else
	strcpy(buff, ".");

    bgm.pathroot = strdup(buff);

    if((fp = fopen(bgmindex, "r")) == NULL)
    {
	perror(bgmindex);
	exit(1);
    }

    if(fgets(buff, BUFSIZ, fp) == NULL)
    {
	fprintf(stderr, "%s: Invalid file format.\n", bgmindex);
	exit(1);
    }

    if((bgm.n = atoi(buff)) <= 0)
    {
	fprintf(stderr, "%s: Invalid file format.\n", bgmindex);
	exit(1);
    }

    bgm.bgm = (struct BGM *)malloc(bgm.n * sizeof(struct BGM));
    for(i = 0; i < bgm.n; i++)
    {
	int j, len;

	if(fgets(buff, BUFSIZ, fp) == NULL)
	{
	    fprintf(stderr, "%s: Invalid file format.\n", bgmindex);
	    exit(1);
	}
	len = strlen(buff);

	if((bgm.bgm[i].n = atoi(buff)) <= 0)
	{
	    fprintf(stderr, "%s: Invalid file format.\n", bgmindex);
	    exit(1);
	}
	buff[len - 1] = '\0';
	bgm.bgm[i].tracks =
	    (struct BGMmap *)malloc(bgm.bgm[i].n * sizeof(struct BGMmap));

	p = buff;
	while(*p == ' ') p++;
	p = strdup(p);
	for(j = 0; j < bgm.bgm[i].n; j++)
	{
	    if((p = strchr(p, ' ')) == NULL)
	    {
		fprintf(stderr, "%s: Invalid file format.\n", bgmindex);
		exit(1);
	    }
	    while(*p == ' ') *p++ = '\0';
	    bgm.bgm[i].tracks[j].name = p;
	}

#ifdef DEBUG
	printf("BGM %d", i);
#endif /* DEBUG */

	for(j = 0; j < bgm.bgm[i].n; j++)
	{
	    int fd;
	    struct stat st;

	    sprintf(buff, "%s/%s", bgm.pathroot, bgm.bgm[i].tracks[j].name);

#ifdef DEBUG
	    printf("\t%s\n", buff);
#endif /* DEBUG */

	    if(stat(buff, &st) < 0)
	    {
		perror(buff);
		exit(1);
	    }
	    if(!S_ISREG(st.st_mode))
	    {
		fprintf(stderr, "%s: Not a regular file.\n", buff);
		exit(1);
	    }
	    if(((S_IRUSR | S_IRGRP | S_IROTH) & st.st_mode) !=
	       (S_IRUSR | S_IRGRP | S_IROTH))
	    {
		fprintf(stderr, "%s: Invalid read permission.\n", buff);
		exit(1);
	    }

	    bgm.bgm[i].tracks[j].mapped = NULL;
	    bgm.bgm[i].tracks[j].size = st.st_size;
	}
    }

    fclose(fp);
}

static void unmap_bgm(struct BGM* b)
{
    int i;

    for(i = 0; i < b->n; i++)
	if(b->tracks[i].mapped != NULL)
	{
#ifdef DEBUG
	    printf("Unmapping %s\n", b->tracks[i].name);
#endif /* DEBUG */
	    munmap(b->tracks[i].mapped, b->tracks[i].size);
	    b->tracks[i].mapped = NULL;
	}
}

static void map_bgm(struct BGM* b)
{
    int i;
    char buff[BUFSIZ];

    for(i = 0; i < b->n; i++)
    {
	if(b->tracks[i].mapped == NULL)
	{
	    int fd;

	    sprintf(buff, "%s/%s", bgm.pathroot, b->tracks[i].name);
#ifdef DEBUG
	    printf("Mapping %s\n", b->tracks[i].name);
#endif /* DEBUG */

	    if((fd = open(buff, O_RDONLY)) < 0)
	    {
		perror(buff);
		continue;
	    }

	    if((b->tracks[i].mapped = mmap(0, b->tracks[i].size, PROT_READ,
					   MAP_PRIVATE, fd, 0)) == (caddr_t)-1)
	    {
		perror(buff);
		b->tracks[i].mapped = NULL;
		continue;
	    }
	}
    }
}

void set_bgm(int stage)
{
    stage = stage % bgm.n;

    if(current_bgm == NULL)
	current_bgm = bgm.bgm + stage;
    else
    {
	static struct BGM* tmp;

	tmp = bgm.bgm + stage;
	if(tmp != current_bgm)
	    unmap_bgm(current_bgm);
	current_bgm = tmp;
    }
    map_bgm(current_bgm);
    current_bgm->cur = 0;
}

char* next_bgm(int* nbytes_ret)
{
    char* p;
    int i;

    if(current_bgm == NULL)
	return NULL;

    i = current_bgm->cur;
    p = current_bgm->tracks[i].mapped;
    *nbytes_ret = current_bgm->tracks[i].size;
    if(i + 1 < current_bgm->n)
	current_bgm->cur++;

    return p;
}    


#if 0
void main(int argc, char** argv)
{
    init_bgm(argv[1]);

    set_bgm(0);
    set_bgm(4);
    set_bgm(4);
    set_bgm(7);
    set_bgm(7);
    set_bgm(7);
    set_bgm(100);
    set_bgm(0);
    exit(0);
}
#endif
